lists.openwall.net   lists  /  announce  owl-users  owl-dev  john-users  john-dev  passwdqc-users  yescrypt  popa3d-users  /  oss-security  kernel-hardening  musl  sabotage  tlsify  passwords  /  crypt-dev  xvendor  /  Bugtraq  Full-Disclosure  linux-kernel  linux-netdev  linux-ext4  linux-hardening  linux-cve-announce  PHC 
Open Source and information security mailing list archives
 
Hash Suite: Windows password security audit tool. GUI, reports in PDF.
[<prev] [next>] [thread-next>] [day] [month] [year] [list]
Date:	Sat, 05 Jul 2008 05:59:30 -0700 (PDT)
From:	leon.woestenberg@...il.com
To:	Leon Woestenberg <leon@...ebranch.com>
Subject: [PATCH 2.6.24.7-rt14 v2] rt: {queue,schedule}_work_prio() allowing work priorities other than caller's priority

Adds schedule_work_prio() and queue_work_prio() so that work can be
scheduled with a different priority than the caller.

Currently under PREEMPT RT, queue_work() and therefore schedule_work() makes
the work inherit the caller's priority.

However, if a real-time kernel interrupt handler thread defers work it may want
the deferred work have lower priority so it does not contend for the CPU with
succeeding interrupt handling. In the non real-time kernel, this was guaranteed
because interrupts preempt workqueues.

The faulty case was a non-FIFO serial driver that deferred LED blinking to
a workqueue using schedule_work(). However, that work used GPIO bitbanged I2C,
which uses 50 usecs udelay()s. With the work inheriting the serial IRQ priority,
it easily missed the ~60 usec deadline of 115200 bps communications.

This patch solves the issue by introducing and using schedule_work_prio().

Signed-off-by: Leon Woestenberg <leon@...ebranch.com>

Index: linux-2.6.24/include/linux/workqueue.h
===================================================================
--- linux-2.6.24.orig/include/linux/workqueue.h	2008-06-30 17:12:57.000000000 +0200
+++ linux-2.6.24/include/linux/workqueue.h	2008-06-30 17:30:34.000000000 +0200
@@ -185,6 +185,7 @@
 extern void destroy_workqueue(struct workqueue_struct *wq);
 
 extern int FASTCALL(queue_work(struct workqueue_struct *wq, struct work_struct *work));
+extern int FASTCALL(queue_work_prio(struct workqueue_struct *wq, struct work_struct *work, int prio));
 extern int FASTCALL(queue_delayed_work(struct workqueue_struct *wq,
 			struct delayed_work *work, unsigned long delay));
 extern int queue_delayed_work_on(int cpu, struct workqueue_struct *wq,
@@ -194,6 +195,7 @@
 extern void flush_scheduled_work(void);
 
 extern int FASTCALL(schedule_work(struct work_struct *work));
+extern int FASTCALL(schedule_work_prio(struct work_struct *work, int prio));
 extern int FASTCALL(schedule_delayed_work(struct delayed_work *work,
 					unsigned long delay));
 extern int schedule_delayed_work_on(int cpu, struct delayed_work *work,
Index: linux-2.6.24/kernel/workqueue.c
===================================================================
--- linux-2.6.24.orig/kernel/workqueue.c	2008-06-30 17:12:05.000000000 +0200
+++ linux-2.6.24/kernel/workqueue.c	2008-06-30 17:25:03.000000000 +0200
@@ -163,9 +163,10 @@
 }
 
 /**
  * queue_work - queue work on a workqueue
  * @wq: workqueue to use
  * @work: work to queue
+ * @prio: priority of work
  *
  * Returns 0 if @work was already on a queue, non-zero otherwise.
  *
@@ -174,17 +175,36 @@
  *
  * Especially no such guarantee on PREEMPT_RT.
  */
-int fastcall queue_work(struct workqueue_struct *wq, struct work_struct *work)
+int fastcall queue_work_prio(struct workqueue_struct *wq,
+  struct work_struct *work, int prio)
 {
 	int ret = 0, cpu = raw_smp_processor_id();
 
 	if (!test_and_set_bit(WORK_STRUCT_PENDING, work_data_bits(work))) {
 		BUG_ON(!plist_node_empty(&work->entry));
-		__queue_work(wq_per_cpu(wq, cpu), work, current->normal_prio);
+		__queue_work(wq_per_cpu(wq, cpu), work, prio);
 		ret = 1;
 	}
 	return ret;
 }
+EXPORT_SYMBOL_GPL(queue_work_prio);
+
+/**
+ * queue_work - queue work on a workqueue. work inherits caller's prio.
+ * @wq: workqueue to use
+ * @work: work to queue
+ *
+ * Returns 0 if @work was already on a queue, non-zero otherwise.
+ *
+ * We queue the work to the CPU it was submitted, but there is no
+ * guarantee that it will be processed by that CPU.
+ *
+ * Especially no such guarantee on PREEMPT_RT.
+ */
+int fastcall queue_work(struct workqueue_struct *wq, struct work_struct *work)
+{
+  return queue_work_prio(wq, work, current->normal_prio);
+}
 EXPORT_SYMBOL_GPL(queue_work);
 
 void delayed_work_timer_fn(unsigned long __data)
@@ -627,7 +647,8 @@
  * schedule_work - put work task in global workqueue
  * @work: job to be done
  *
- * This puts a job in the kernel-global workqueue.
+ * This puts a job in the kernel-global workqueue. The job inherits the
+ * caller's priority.
  */
 int fastcall schedule_work(struct work_struct *work)
 {
@@ -636,6 +657,19 @@
 EXPORT_SYMBOL(schedule_work);
 
 /**
+ * schedule_work_prio - put work task in global workqueue, set work prio
+ * @work: job to be done
+ *
+ * This puts a job in the kernel-global workqueue. The job gets the specified
+ * priority.
+ */
+int fastcall schedule_work_prio(struct work_struct *work, int prio)
+{
+	return queue_work_prio(keventd_wq, work, prio);
+}
+EXPORT_SYMBOL(schedule_work_prio);
+
+/**
  * schedule_delayed_work - put work task in global workqueue after delay
  * @dwork: job to be done
  * @delay: number of jiffies to wait or 0 for immediate execution
@@ -163,9 +163,10 @@
 
 /**
- * queue_work - queue work on a workqueue
+ * queue_work_prio - queue work on a workqueue and set work priority
  * @wq: workqueue to use
  * @work: work to queue
--
To unsubscribe from this list: send the line "unsubscribe linux-kernel" in
the body of a message to majordomo@...r.kernel.org
More majordomo info at  http://vger.kernel.org/majordomo-info.html
Please read the FAQ at  http://www.tux.org/lkml/

Powered by blists - more mailing lists

Powered by Openwall GNU/*/Linux Powered by OpenVZ